@@ -8,11 +8,11 @@ test_that("Contourmap, test ind", {
   ind3[1:4] = TRUE
 
   res1 = contourmap(data$mu, data$Q, n.levels=2, ind=ind1,
-                    seed=data$seed, alpha=0.1)
+                    seed=data$seed, alpha=0.1,max.threads=1)
   res2 = contourmap(data$mu, data$Q, n.levels=2, ind=ind2,
-                    seed=data$seed, alpha=0.1)
+                    seed=data$seed, alpha=0.1,max.threads=1)
   res3 = contourmap(data$mu, data$Q, n.levels=2, ind=ind3,
-                    seed=data$seed, alpha=0.1)
+                    seed=data$seed, alpha=0.1,max.threads=1)
 
   expect_equal(res1$F,res2$F,tolerance=1e-7)
   expect_equal(res2$F,res3$F,tolerance=1e-7)
